"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.migrateSchema = exports.diffSchema = void 0;
const clc = require("colorette");
const sql_formatter_1 = require("sql-formatter");
const types_1 = require("./types");
const client_1 = require("./client");
const connect_1 = require("../gcp/cloudsql/connect");
const prompt_1 = require("../prompt");
const logger_1 = require("../logger");
const error_1 = require("../error");
const projectUtils_1 = require("../projectUtils");
const utils_1 = require("../utils");
const experiments = require("../experiments");
const errors = require("./errors");
async function diffSchema(schema) {
    const { serviceName, instanceName, databaseId } = getIdentifiers(schema);
    await ensureServiceIsConnectedToCloudSql(serviceName, instanceName, databaseId, false);
    setCompatibleMode(schema, databaseId, instanceName);
    try {
        await (0, client_1.upsertSchema)(schema, true);
        (0, utils_1.logLabeledSuccess)("dataconnect", `Database schema is up to date.`);
    }
    catch (err) {
        if (err.status !== 400) {
            throw err;
        }
        const invalidConnectors = errors.getInvalidConnectors(err);
        const incompatible = errors.getIncompatibleSchemaError(err);
        if (!incompatible && !invalidConnectors.length) {
            throw err;
        }
        if (invalidConnectors.length) {
            displayInvalidConnectors(invalidConnectors);
        }
        if (incompatible) {
            displaySchemaChanges(incompatible);
            return incompatible.diffs;
        }
    }
    return [];
}
exports.diffSchema = diffSchema;
async function migrateSchema(args) {
    const { options, schema, validateOnly } = args;
    const { serviceName, instanceId, instanceName, databaseId } = getIdentifiers(schema);
    await ensureServiceIsConnectedToCloudSql(serviceName, instanceName, databaseId, true);
    setCompatibleMode(schema, databaseId, instanceName);
    try {
        await (0, client_1.upsertSchema)(schema, validateOnly);
        logger_1.logger.debug(`Database schema was up to date for ${instanceId}:${databaseId}`);
    }
    catch (err) {
        if (err.status !== 400) {
            throw err;
        }
        const incompatible = errors.getIncompatibleSchemaError(err);
        const invalidConnectors = errors.getInvalidConnectors(err);
        if (!incompatible && !invalidConnectors.length) {
            throw err;
        }
        const migrationMode = await promptForSchemaMigration(options, databaseId, incompatible, validateOnly);
        const shouldDeleteInvalidConnectors = await promptForInvalidConnectorError(options, serviceName, invalidConnectors, validateOnly);
        let diffs = [];
        if (incompatible) {
            diffs = await handleIncompatibleSchemaError({
                options,
                databaseId,
                instanceId,
                incompatibleSchemaError: incompatible,
                choice: migrationMode,
            });
        }
        if (shouldDeleteInvalidConnectors) {
            await deleteInvalidConnectors(invalidConnectors);
        }
        if (!validateOnly) {
            await (0, client_1.upsertSchema)(schema, validateOnly);
        }
        return diffs;
    }
    return [];
}
exports.migrateSchema = migrateSchema;
function setCompatibleMode(schema, databaseId, instanceName) {
    var _a;
    if (experiments.isEnabled("fdccompatiblemode")) {
        if ((_a = schema.primaryDatasource.postgresql) === null || _a === void 0 ? void 0 : _a.schemaValidation) {
            schema.primaryDatasource.postgresql.schemaValidation = "COMPATIBLE";
        }
        else {
            schema.primaryDatasource = {
                postgresql: {
                    database: databaseId,
                    cloudSql: {
                        instance: instanceName,
                    },
                    schemaValidation: "COMPATIBLE",
                },
            };
        }
    }
}
function getIdentifiers(schema) {
    var _a, _b;
    const databaseId = (_a = schema.primaryDatasource.postgresql) === null || _a === void 0 ? void 0 : _a.database;
    if (!databaseId) {
        throw new error_1.FirebaseError("Schema is missing primaryDatasource.postgresql?.database, cannot migrate");
    }
    const instanceName = (_b = schema.primaryDatasource.postgresql) === null || _b === void 0 ? void 0 : _b.cloudSql.instance;
    if (!instanceName) {
        throw new error_1.FirebaseError("tried to migrate schema but instance name was not provided in dataconnect.yaml");
    }
    const instanceId = instanceName.split("/").pop();
    const serviceName = schema.name.replace(`/schemas/${types_1.SCHEMA_ID}`, "");
    return {
        databaseId,
        instanceId,
        instanceName,
        serviceName,
    };
}
function suggestedCommand(serviceName, invalidConnectorNames) {
    const serviceId = serviceName.split("/")[5];
    const connectorIds = invalidConnectorNames.map((i) => i.split("/")[7]);
    const onlys = connectorIds.map((c) => `dataconnect:${serviceId}:${c}`).join(",");
    return `firebase deploy --only ${onlys}`;
}
async function handleIncompatibleSchemaError(args) {
    const { incompatibleSchemaError, options, instanceId, databaseId, choice } = args;
    if (incompatibleSchemaError.destructive && choice === "safe") {
        throw new error_1.FirebaseError("This schema migration includes potentially destructive changes. If you'd like to execute it anyway, rerun this command with --force");
    }
    const projectId = (0, projectUtils_1.needProjectId)(options);
    const iamUser = await (0, connect_1.setupIAMUser)(instanceId, databaseId, options);
    const commandsToExecute = incompatibleSchemaError.diffs
        .filter((d) => {
        switch (choice) {
            case "all":
                return true;
            case "safe":
                return !d.destructive;
            case "none":
                return false;
        }
    })
        .map((d) => d.sql);
    if (commandsToExecute.length) {
        await (0, connect_1.execute)([
            `SET ROLE "${(0, connect_1.firebaseowner)(databaseId)}"`,
            ...commandsToExecute,
            `GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA "public" TO PUBLIC`,
        ], {
            projectId,
            instanceId,
            databaseId,
            username: iamUser,
        });
        return incompatibleSchemaError.diffs;
    }
    return [];
}
async function promptForSchemaMigration(options, databaseName, err, validateOnly) {
    if (!err) {
        return "none";
    }
    displaySchemaChanges(err);
    if (!options.nonInteractive) {
        if (validateOnly && options.force) {
            return "all";
        }
        const choices = err.destructive
            ? [
                { name: "Execute all changes (including destructive changes)", value: "all" },
                { name: "Abort changes", value: "none" },
            ]
            : [
                { name: "Execute changes", value: "all" },
                { name: "Abort changes", value: "none" },
            ];
        return await (0, prompt_1.promptOnce)({
            message: `Would you like to execute these changes against ${databaseName}?`,
            type: "list",
            choices,
        });
    }
    if (!validateOnly) {
        throw new error_1.FirebaseError("Command aborted. Your database schema is incompatible with your Data Connect schema. Run `firebase dataconnect:sql:migrate` to migrate your database schema");
    }
    else if (options.force) {
        return "all";
    }
    else if (!err.destructive) {
        return "all";
    }
    else {
        throw new error_1.FirebaseError("Command aborted. This schema migration includes potentially destructive changes. If you'd like to execute it anyway, rerun this command with --force");
    }
}
async function promptForInvalidConnectorError(options, serviceName, invalidConnectors, validateOnly) {
    if (!invalidConnectors.length) {
        return false;
    }
    displayInvalidConnectors(invalidConnectors);
    if (validateOnly) {
        if (options.force) {
            return false;
        }
        throw new error_1.FirebaseError(`Command aborted. If you'd like to migrate it anyway, you may override with --force.`);
    }
    if (options.force) {
        return true;
    }
    if (!options.nonInteractive &&
        (await (0, prompt_1.confirm)(Object.assign(Object.assign({}, options), { message: `Would you like to delete and recreate these connectors? This will cause ${clc.red(`downtime.`)}.` })))) {
        return true;
    }
    const cmd = suggestedCommand(serviceName, invalidConnectors);
    throw new error_1.FirebaseError(`Command aborted. Try deploying those connectors first with ${clc.bold(cmd)}`);
}
async function deleteInvalidConnectors(invalidConnectors) {
    return Promise.all(invalidConnectors.map(client_1.deleteConnector));
}
function displayInvalidConnectors(invalidConnectors) {
    const connectorIds = invalidConnectors.map((i) => i.split("/").pop()).join(", ");
    (0, utils_1.logLabeledWarning)("dataconnect", `The schema you are deploying is incompatible with the following existing connectors: ${connectorIds}.`);
    (0, utils_1.logLabeledWarning)("dataconnect", `This is a ${clc.red("breaking")} change and may break existing apps.`);
}
async function ensureServiceIsConnectedToCloudSql(serviceName, instanceId, databaseId, linkIfNotConnected) {
    let currentSchema = await (0, client_1.getSchema)(serviceName);
    if (!currentSchema) {
        if (!linkIfNotConnected) {
            (0, utils_1.logLabeledWarning)("dataconnect", `Not yet linked to the Cloud SQL instance.`);
            return;
        }
        (0, utils_1.logLabeledBullet)("dataconnect", `Linking the Cloud SQL instance...`);
        currentSchema = {
            name: `${serviceName}/schemas/${types_1.SCHEMA_ID}`,
            source: {
                files: [],
            },
            primaryDatasource: {
                postgresql: {
                    database: databaseId,
                    cloudSql: {
                        instance: instanceId,
                    },
                },
            },
        };
    }
    const postgresql = currentSchema.primaryDatasource.postgresql;
    if ((postgresql === null || postgresql === void 0 ? void 0 : postgresql.cloudSql.instance) !== instanceId) {
        (0, utils_1.logLabeledWarning)("dataconnect", `Switching connected Cloud SQL instance\nFrom ${postgresql === null || postgresql === void 0 ? void 0 : postgresql.cloudSql.instance}\nTo ${instanceId}`);
    }
    if ((postgresql === null || postgresql === void 0 ? void 0 : postgresql.database) !== databaseId) {
        (0, utils_1.logLabeledWarning)("dataconnect", `Switching connected Postgres database from ${postgresql === null || postgresql === void 0 ? void 0 : postgresql.database} to ${databaseId}`);
    }
    if (!postgresql || postgresql.schemaValidation === "STRICT") {
        return;
    }
    postgresql.schemaValidation = "STRICT";
    try {
        await (0, client_1.upsertSchema)(currentSchema, false);
    }
    catch (err) {
        if (err.status >= 500) {
            throw err;
        }
        logger_1.logger.debug(err);
    }
}
function displaySchemaChanges(error) {
    switch (error.violationType) {
        case "INCOMPATIBLE_SCHEMA":
            {
                const message = "Your new schema is incompatible with the schema of your CloudSQL database. " +
                    "The following SQL statements will migrate your database schema to match your new Data Connect schema.\n" +
                    error.diffs.map(toString).join("\n");
                (0, utils_1.logLabeledWarning)("dataconnect", message);
            }
            break;
        case "INACCESSIBLE_SCHEMA":
            {
                const message = "Cannot access your CloudSQL database to validate schema. " +
                    "The following SQL statements can setup a new database schema.\n" +
                    error.diffs.map(toString).join("\n");
                (0, utils_1.logLabeledWarning)("dataconnect", message);
                (0, utils_1.logLabeledWarning)("dataconnect", "Some SQL resources may already exist.");
            }
            break;
        default:
            throw new error_1.FirebaseError(`Unknown schema violation type: ${error.violationType}, IncompatibleSqlSchemaError: ${error}`);
    }
}
function toString(diff) {
    return `\/** ${diff.destructive ? clc.red("Destructive: ") : ""}${diff.description}*\/\n${(0, sql_formatter_1.format)(diff.sql, { language: "postgresql" })}`;
}
